export EulerEq
module EulerEq


using dg1d
using .EquationOfState
using LoopVectorization
using Random


include("types.jl")
include("equation.jl")
include("rhs.jl")
include("initialdata.jl")
include("callbacks.jl")
include("setup.jl")


#######################################################################
#                 dg1d project interface definitions                  #
#######################################################################


dg1d.project_type(::Val{:EulerEq}) = Project

# TODO Remove bdryconds
dg1d.rhs!(env::Environment, P, bdryconds) = rhs!(env, P, P.hrsc, env.mesh)

dg1d.timestep!(env::Environment, P::Union{Project,Project2d}) = timestep(env, P, P.hrsc, env.mesh)

dg1d.@parameters EulerEq begin

  """
  Initial data configuration. Available options:
  - `smooth_isentropic_flow`
  - `sod_shock_tube`
  - `shu_osher_shock_tube_v1`
  - `shu_osher_shock_tube_v2`
  - `kelvin_helmholtz`: Requires 2d mesh.
  - `blast_wave`
  """
  id = "sine"
  @check id in [ "smooth_isentropic_flow", "sod_shock_tube",
                 "shu_osher_shock_tube_v1", "shu_osher_shock_tube_v2",
                 "kelvin_helmholtz",
                 "blast_wave" ]

  """
  If `true` smooth initial data using a Bernstein filter.
  """
  id_smooth = false
  @check id_smooth isa Bool

  """
  Boundary conditions. Available options:
  - `periodic`
  - `from_id`
  """
  bc = "periodic"
  @check bc in [ "periodic", "from_id" ]

  """
  Type of artificial viscosity regularization. Available options:
  - `none`
  - `mono`: monolithic regularization, see [1]
  - `navierstokes`: regularization based on Navier-Stokes equations, see [1]; actually implemented version taken from [3]
    Also see parameter `av_navierstokes_prho, av_navierstokes_pt`.
  - `general`: most general regularization, see [1]; actually implemented version taken from [2]

  [1] Guermond+, 2014
  [2] Guermond+, 2016
  [3] Zingan+, 2013
  """
  av_regularization = "none"
  @check av_regularization in ["none", "mono", "navierstokes", "general"]

  """
  Prandtl number P_ρ to compute the viscosity ν for the density regularization
  from the the viscosity μ of the momentum regularization, eg. ν = P_ρ μ.
  See Zingan+, 2013.
  """
  av_navierstokes_prho = 1.0
  @check av_navierstokes_prho ≥ 0.0

  """
  Prandtl number P_T to compute the viscosity κ for the energy regularization
  from the the viscosity μ of the momentum regularization, eg. κ = P_T μ.
  See Zingan+, 2013.
  """
  av_navierstokes_pt = 1.0
  @check av_navierstokes_pt ≥ 0.0

  """
  Determines how quickly artificial viscosity can alter between time steps.
  If `av_drag = 0.0` then no dragging occurs.
  If `av_drag = 1.0` then viscosity is frozen to its initial value.
  """
  av_drag = 0.0
  @check 0.0 <= av_drag <= 1.0

  """
  If `true` any artificial viscosity related callbacks are run on every substep of the evolution.
  """
  av_recompute_substeps = false
  @check av_recompute_substeps isa Bool

  """
  The HRSC method used for FV discretization.
  Note: For DG methods use the `HRSC` parameter section.
  - `fv_central` ... Riemann solver free FV method, cf. Hesthaven book 2018, sec. 6.3
  - `slope_limiter` ... slope limited HRSC, cf. Hesthaven book 2018, sec. 10.2
  """
  hrsc = "fv_central"
  @check hrsc in [ "fv_central", "slope_limiter" ]

  """
  Available options:
  - `none`
  - `minmod` [1]
  - `muscl` [1] ... monotone upstream-centered limiter
  - `superbee` [1]
  - `van_albada` [1]
  - `van_leer` [1]
  - `tvb` [1] ... total variation bounded

  [1] Hesthaven 2018 book, p. 268, Script 10.10
  """
  slope_limiter_method = "muscl"
  @check slope_limiter_method in [ "none", "minmod", "muscl", "superbee", "van_leer", "tvb" ]

  """
  Slope limiting parameter for MUSCL scheme, cf. Hesthaven book 2008, sec 10.2
  """
  slope_limiter_tvb_M = 150.0
  @check slope_limiter_tvb_M > 0.0

end

end # module
