#######################################################################
#                              Timestep                               #
#######################################################################

function compute_maxspeed(mesh, equation, cache)
  @unpack v = get_static_variables(cache)
  broadcast_volume!(maxspeed, equation, mesh)
  vmax = dg1d.absolute_maximum(v)
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end

function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh1d)
  vmax = compute_maxspeed(mesh, P.equation, mesh.cache)
  dl = min_grid_spacing(mesh)
  dt = dl / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = mesh.element.N


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(maxspeed, equation, mesh)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack element = mesh
  @unpack N = element

  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K
  dl = min_grid_spacing(mesh)

  N = dtfactor(mesh)
  dt1 = dl / vmax
  dt2 = dL^2 / (mumax * N^3)
  dt = min(dt1, dt2)

  return dt
end


function timestep(env, P::Project2d, hrsc::Maybe{HRSC.AbstractHRSC}, ::Mesh2d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(maxspeed, equation, mesh)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = dl / vmax

  return dt
end


#######################################################################
#                               FV rhs                                #
#######################################################################

function rhs!(env, P::Project, hrsc::Nothing, mesh::Mesh1d{FVElement})
  if P.prms.hrsc === :fv_central
    rhs_fv_central!(P, mesh)
  elseif P.prms.hrsc === :slope_limiter
    rhs_fv_muscl!(P, mesh)
  else
    TODO(P.prms.hrsc)
  end
end


function rhs_fv_central!(P::Project, mesh::Mesh1d{FVElement})

  @unpack equation                 = P
  @unpack rho, q, E                = get_dynamic_variables(mesh)
  @unpack flx_rho, flx_q, flx_E    = get_static_variables(mesh)
  @unpack rhs_rho, rhs_q, rhs_E    = get_rhs_variables(mesh)
  @unpack nflx_rho, nflx_q, nflx_E = get_bdry_variables(mesh)
  @unpack bdry_rho, bdry_q, bdry_E = get_bdry_variables(mesh)
  @unpack dt                       = get_global_variables(mesh)

  broadcast_volume!(flux, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_rho, rho, flx_rho, bdry_rho, nflx_rho, dt[1], mesh)
  fv_update_step!(rhs_q,   q,   flx_q,   bdry_q,   nflx_q,   dt[1], mesh)
  fv_update_step!(rhs_E,   E,   flx_E,   bdry_E,   nflx_E,   dt[1], mesh)

  return
end


function rhs_fv_muscl!(P::Project, mesh::Mesh1d{FVElement})
  @unpack equation              = P
  @unpack rho, q, E             = get_dynamic_variables(mesh)
  @unpack rho1, q1, E1,
          rho2, q2, E2          = get_static_variables(mesh)
  @unpack rhs_rho, rhs_q, rhs_E = get_rhs_variables(mesh)
  @unpack dt = get_global_variables(mesh)
  dt = dt[1]

  # SSPRK3 stepping from Hesthaven 2018 book, sec 10.2 (cf Script 10.13)
  step_fv_muscl!(P,mesh,(rho,q,E))
  @. rho1 = rho + dt*rhs_rho
  @. q1   = q   + dt*rhs_q
  @. E1   = E   + dt*rhs_E
  step_fv_muscl!(P,mesh,(rho1,q1,E1))
  @. rho2 = (3*rho + rho1 + dt*rhs_rho)/4
  @. q2   = (3*q   + q1   + dt*rhs_q)/4
  @. E2   = (3*E   + E1   + dt*rhs_E)/4
  step_fv_muscl!(P,mesh,(rho2,q2,E2))
  # atm final result is published via rhs_ variables
  @. rhs_rho  = (rho + 2*rho2 + 2*dt*rhs_rho)/3
  @. rhs_q    = (q   + 2*q2   + 2*dt*rhs_q)/3
  @. rhs_E    = (E   + 2*E2   + 2*dt*rhs_E)/3

  return
end


function step_fv_muscl!(P::Project, mesh::Mesh1d{FVElement}, state)

  rho, q, E = state

  @unpack equation                    = P
  @unpack rhs_rho, rhs_q, rhs_E       = get_rhs_variables(mesh)
  @unpack rhoL, qL, EL, pL,
          rhoR, qR, ER, pR,
          drho_p, dq_p, dE_p,
          drho_m, dq_m, dE_m,
          flx_rho_L, flx_q_L, flx_E_L,
          flx_rho_R, flx_q_R, flx_E_R,
          nflx_rho_L, nflx_q_L, nflx_E_L,
          nflx_rho_R, nflx_q_R, nflx_E_R,
          v                           = get_static_variables(mesh)
  @unpack eos                         = equation

  periodic = isperiodic(mesh)
  prms_slope_limiter = (; method=P.prms.slope_limiter_method,
                          tvb_M=P.prms.slope_limiter_tvb_M)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)
  dl = L/K

  for k in 1:K
    eps = E[k] / rho[k] - q[k]^2 / (2*rho[k]^2)
    p = eos(Pressure, Density, rho[k], InternalEnergy, eps)
    v[k] = abs(q[k]/p) + sqrt(abs(eos.gamma * p / rho[k]))
  end

  for vars in ( (rhoL,rhoR,rho,drho_m,drho_p),
                (qL,qR,q,dq_m,dq_p),
                (EL,ER,E,dE_m,dE_p) )

    vL,vR,v,dv₋,dv₊ = vars

    # compute left (-) and right (+) differences
    @turbo for k in 2:K-1
      k₋, k₊ = k-1, k+1
      v₋, v₀, v₊ = v[k₋], v[k], v[k₊]
      dv₊[k], dv₋[k] = v₊-v₀, v₀-v₋
    end
    if periodic
      for k in (1,K)
        k₋, k₊ = periodic_index(k-1,K), periodic_index(k+1,K)
        v₋, v₀, v₊ = v[k₋], v[k], v[k₊]
        dv₊[k], dv₋[k] = v₊-v₀, v₀-v₋
      end
    else
      # TODO Atm we assume that evolution does not reach bdry
    end

    # limit slope and compute left (L) and (R) interface values
    for k in 1:K
      dvL = limit_slope(dv₊[k], dv₋[k], dl, prms_slope_limiter)
      vL[k] = v[k] - dvL/2
      vR[k] = v[k] + dvL/2
    end
  end

  for k in 1:K
    epsL = EL[k] / rhoL[k] - qL[k]^2 / (2*rhoL[k]^2)
    pL[k] = eos(Pressure, Density, rhoL[k], InternalEnergy, epsL)
    epsR = ER[k] / rhoR[k] - qR[k]^2 / (2*rhoR[k]^2)
    pR[k] = eos(Pressure, Density, rhoR[k], InternalEnergy, epsR)
  end

  # numerical fluxes
  # subscript + ... cell interior value
  # subscript - ... cell exterior value
  # nx ... pointing from interior too exterior cell
  @unpack v = get_static_variables(mesh)
  for k in 2:K
    # on left cell interfaces
    nx            = -1
    k₊, k₋        = k, k-1
    v₊, v₋        = v[k₊], v[k₋]
    rho₊, rho₋    = rhoL[k₊], rhoR[k₋]
    q₊, q₋        = qL[k₊], qR[k₋]
    E₊, E₋        = EL[k₊], ER[k₋]
    p₊, p₋        = pL[k₊], pR[k₋]
    max_v         = absmax(v₊, v₋)
    f₊, f₋        = q₊, q₋
    nflx_rho_L[k] = LLF(nx*f₊, nx*f₋, rho₊, rho₋, max_v)
    f₊, f₋        = q₊^2/rho₊+p₊, q₋^2/rho₋+p₋
    nflx_q_L[k]   = LLF(nx*f₊, nx*f₋, q₊, q₋, max_v)
    f₊, f₋        = (E₊+p₊)*q₊/rho₊, (E₋+p₋)*q₋/rho₋
    nflx_E_L[k]   = LLF(nx*f₊, nx*f₋, E₊, E₋, max_v)
  end
  for k in 1:K-1
    # on right cell interfaces
    nx            = 1
    k₊, k₋        = k, k+1
    v₊, v₋        = v[k₊], v[k₋]
    rho₊, rho₋    = rhoR[k₊], rhoL[k₋]
    q₊, q₋        = qR[k₊], qL[k₋]
    E₊, E₋        = ER[k₊], EL[k₋]
    p₊, p₋        = pR[k₊], pL[k₋]
    max_v         = absmax(v₊, v₋)
    f₊, f₋        = q₊, q₋
    nflx_rho_R[k] = LLF(nx*f₊, nx*f₋, rho₊, rho₋, max_v)
    f₊, f₋        = q₊^2/rho₊+p₊, q₋^2/rho₋+p₋
    nflx_q_R[k]   = LLF(nx*f₊, nx*f₋, q₊, q₋, max_v)
    f₊, f₋        = (E₊+p₊)*q₊/rho₊, (E₋+p₋)*q₋/rho₋
    nflx_E_R[k]   = LLF(nx*f₊, nx*f₋, E₊, E₋, max_v)
  end

  # bdry conditions
  if periodic
    begin
      k             = 1
      nx            = -1
      k₊, k₋        = periodic_index(k,K), periodic_index(k-1,K)
      v₊, v₋        = v[k₊], v[k₋]
      rho₊, rho₋    = rhoL[k₊], rhoR[k₋]
      q₊, q₋        = qL[k₊], qR[k₋]
      E₊, E₋        = EL[k₊], ER[k₋]
      p₊, p₋        = pL[k₊], pR[k₋]
      max_v         = absmax(v₊, v₋)
      f₊, f₋        = q₊, q₋
      nflx_rho_L[k] = LLF(nx*f₊, nx*f₋, rho₊, rho₋, max_v)
      f₊, f₋        = q₊^2/rho₊+p₊, q₋^2/rho₋+p₋
      nflx_q_L[k]   = LLF(nx*f₊, nx*f₋, q₊, q₋, max_v)
      f₊, f₋        = (E₊+p₊)*q₊/rho₊, (E₋+p₋)*q₋/rho₋
      nflx_E_L[k]   = LLF(nx*f₊, nx*f₋, E₊, E₋, max_v)
    end
    begin
      k             = K
      nx            = 1
      k₊, k₋        = periodic_index(k,K), periodic_index(k+1,K)
      v₊, v₋        = v[k₊], v[k₋]
      rho₊, rho₋    = rhoR[k₊], rhoL[k₋]
      q₊, q₋        = qR[k₊], qL[k₋]
      E₊, E₋        = ER[k₊], EL[k₋]
      p₊, p₋        = pR[k₊], pL[k₋]
      max_v         = absmax(v₊, v₋)
      f₊, f₋        = q₊, q₋
      nflx_rho_R[k] = LLF(nx*f₊, nx*f₋, rho₊, rho₋, max_v)
      f₊, f₋        = q₊^2/rho₊+p₊, q₋^2/rho₋+p₋
      nflx_q_R[k]   = LLF(nx*f₊, nx*f₋, q₊, q₋, max_v)
      f₊, f₋        = (E₊+p₊)*q₊/rho₊, (E₋+p₋)*q₋/rho₋
      nflx_E_R[k]   = LLF(nx*f₊, nx*f₋, E₊, E₋, max_v)
    end
  else
    # TODO Atm we assume that evolution does not reach bdry
  end

  # evaluate rhs
  @turbo for k in 1:K
    # we have a + in the braces, because we use the nx with fluxes above
    rhs_rho[k] = -(nflx_rho_R[k] + nflx_rho_L[k])/dl
    rhs_q[k]   = -(nflx_q_R[k]   + nflx_q_L[k]  )/dl
    rhs_E[k]   = -(nflx_E_R[k]   + nflx_E_L[k]  )/dl
  end
  if !periodic
    # TODO Atm we assume that evolution does not reach bdry
    rhs_rho[1] = rhs_rho[end] = 0
    rhs_q[1]   = rhs_q[end]   = 0
    rhs_E[1]   = rhs_E[end]   = 0
  end

  return
end


@inline function limit_slope(a, b, dx, prms)
  if prms.method === :none
    return 0.0
  elseif prms.method === :minmod
    return TCI.minmod(a, b)
  elseif prms.method === :muscl
    return TCI.minmod((a+b)/2, 2*a, 2*b)
  elseif prms.method === :superbee
    return TCI.minmod(TCI.maxmod(a,b),TCI.minmod(2*a,2*b))
  elseif prms.method === :van_albada
    c = dx^3
    return TCI.minmod( ((a^2+c^2)*b+(b^2+c^2)*a)/(a^2+b^2+c^2), 2*a, 2*b )
  elseif prms.method === :van_leer
    return TCI.minmod(2*a*b/(a+b), 2*a, 2*b)
  elseif prms.method === :tvb
    return TCI.minmod_M(dx, prms.tvb_M, (a+b)/2, 2*a, 2*b)
  else
    TODO(prms.method)
  end
end


#######################################################################
#                               DG rhs                                #
#######################################################################



function rhs!(env, P::Project, hrsc::Nothing, ::Mesh1d)

  @unpack cache, mesh                   = env
  @unpack equation                      = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q , nflx_E     = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v                        = get_bdry_variables(mesh.cache)
  @unpack v                             = get_static_variables(mesh.cache)

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project2d, hrsc::Nothing, ::Mesh2d)

  @unpack cache, mesh                        = env
  @unpack equation                           = P
  @unpack rho, qx, qy, E                     = get_dynamic_variables(cache)
  @unpack flx_rho_x, flx_rho_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_E_x, flx_E_y, p, eps           = get_static_variables(cache)
  @unpack rhs_rho, rhs_qx, rhs_qy, rhs_E     = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E = get_bdry_variables(cache)

  @unpack bdry_rho, bdry_qx, bdry_qy, bdry_E,
          bdry_v                             = get_bdry_variables(mesh.cache)
  @unpack v                                  = get_static_variables(mesh.cache)

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, qx, bdry_qx)
  dg1d.interpolate_face_data!(mesh, qy, bdry_qy)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho_x, flx_rho_y,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_qx,  flx_qx_x,  flx_qx_y,   nflx_qx,  mesh)
  compute_rhs_weak_form!(rhs_qy,  flx_qy_x,  flx_qy_y,   nflx_qy,  mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E_x,   flx_E_y,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::AbstractReconstruction, ::Mesh1d)

  @unpack cache, mesh                   = env
  @unpack equation                      = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack flag                          = get_cell_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E      = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v                        = get_bdry_variables(mesh.cache)
  @unpack v                             = get_static_variables(mesh.cache)

  isperiodic(x) = true
  HRSC.reconstruct!(rho,  flag, hrsc, isperiodic=isperiodic(mesh))
  HRSC.reconstruct!(q,    flag, hrsc, isperiodic=isperiodic(mesh))
  HRSC.reconstruct!(E,    flag, hrsc, isperiodic=isperiodic(mesh))

  broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  broadcast_faces!(llf, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d)

  @unpack cache, mesh = env
  @unpack equation, prms = P
  reg = prms.av_regularization

  if reg === :mono
    rhs_mono!(cache, mesh, equation, P)
  elseif reg === :navierstokes
    rhs_navierstokes!(cache, mesh, equation, P)
  elseif reg === :general
    rhs_general!(cache, mesh, equation, P)
  else
    TODO(reg)
  end
end


function rhs_mono!(cache, mesh, equation, P)

  @unpack rho, q, E                = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_q, ldg_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E    = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v,
          bdry_ldg_rho, bdry_ldg_q, bdry_ldg_E,
          bdry_smoothed_mu         = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu           = get_static_variables(mesh.cache)

  broadcast_volume!(pressure, equation, mesh)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q, bdry_q)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux, equation, mesh)
  broadcast_bdry!(ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_rho, rho, nflx_rho, mesh)
  compute_rhs_weak_form!(ldg_q,   q,   nflx_q,   mesh)
  compute_rhs_weak_form!(ldg_E,   E,   nflx_E,   mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho, bdry_ldg_rho)
  dg1d.interpolate_face_data!(mesh, ldg_q, bdry_ldg_q)
  dg1d.interpolate_face_data!(mesh, ldg_E, bdry_ldg_E)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_volume!(mono_av_flux, equation, mesh)
  broadcast_faces!(mono_av_nflux, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)
  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end
function rhs_navierstokes!(cache, mesh, equation, P)

  @unpack rho, q, E                        = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, u, e,
          ldg_rho, ldg_u, ldg_e            = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E            = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E, nflx_u,
          nflx_e                           = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v, bdry_u, bdry_e,
          bdry_ldg_rho, bdry_ldg_u, bdry_ldg_e,
          bdry_smoothed_mu                 = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                   = get_static_variables(mesh.cache)
  @unpack mu = get_cell_variables(mesh.cache)

  broadcast_volume!(pressure, equation, mesh)
  broadcast_volume!(navierstokes_vars, equation, mesh)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, q,   bdry_q)
  dg1d.interpolate_face_data!(mesh, E,   bdry_E)
  dg1d.interpolate_face_data!(mesh, v,   bdry_v)
  dg1d.interpolate_face_data!(mesh, u,   bdry_u)
  dg1d.interpolate_face_data!(mesh, e,   bdry_e)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(navierstokes_ldg_nflux, equation, mesh)
  # broadcast_bdry!(navierstokes_ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_rho, rho, nflx_rho, mesh)
  compute_rhs_weak_form!(ldg_u,   u,   nflx_u, mesh)
  compute_rhs_weak_form!(ldg_e,   e,   nflx_e, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho,     bdry_ldg_rho)
  dg1d.interpolate_face_data!(mesh, ldg_u,       bdry_ldg_u)
  dg1d.interpolate_face_data!(mesh, ldg_e,       bdry_ldg_e)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)
  broadcast_volume!(navierstokes_av_flux, equation, mesh)
  broadcast_faces!(navierstokes_av_nflux, equation, mesh)
  broadcast_bdry!(bdryllf, equation, mesh)
  # broadcast_bdry!(navierstokes_av_nflux, equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end
function rhs_general!(cache, mesh, equation)

  TODO()

  @unpack rho, q, E                            = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_rhoe, ldg_u,
          flx_ldg_rho, flx_ldg_q, flx_ldg_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E                = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E,
          nflx_ldg_rho, nflx_ldg_q, nflx_ldg_E = get_bdry_variables(cache)
  @unpack bdry_rho, bdry_q, bdry_E,
          bdry_v,
          bdry_ldg_rho, bdry_ldg_rhoe, bdry_ldg_u,
          bdry_smoothed_mu                     = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                       = get_static_variables(mesh.cache)

  broadcast_volume!(mono_rhoeps, equation, mesh)
  dg1d.interpolate_face_data!(mesh, rho,    bdry_rho)
  dg1d.interpolate_face_data!(mesh, rhoeps, bdry_rhoeps)
  dg1d.interpolate_face_data!(mesh, u,      bdry_u)
  dg1d.interpolate_face_data!(mesh, v,      bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(general_ldg_nflux, equation, mesh)
  broadcast_bdry!(general_ldg_bdryllf, equation, mesh)
  compute_rhs_weak_form!(ldg_rho, rho,    nflx_rho,    mesh)
  compute_rhs_weak_form!(ldg_q,   rhoeps, nflx_rhoeps, mesh)
  compute_rhs_weak_form!(ldg_E,   u,      nflx_u,      mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho,    bdry_ldg_rho)
  dg1d.interpolate_face_data!(mesh, ldg_rhoeps, bdry_ldg_rhoeps)
  dg1d.interpolate_face_data!(mesh, ldg_u,      bdry_ldg_u)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_volume!(general_av_flux, equation, mesh)
  broadcast_faces!(general_av_nflux, equation, mesh)
  broadcast_bdry!(general_bdryllf, equation, mesh)
  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project2d, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh2d)

  @unpack equation                       = P
  @unpack rho, qx, qy, E                 = get_dynamic_variables(mesh.cache)
  @unpack rhs_rho, rhs_qx, rhs_qy, rhs_E = get_rhs_variables(mesh.cache)
  @unpack flx_rho_x, flx_rho_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_E_x, flx_E_y,
          ldg_rho_x, ldg_rho_y,
          ldg_qx_x, ldg_qx_y,
          ldg_qy_x, ldg_qy_y,
          ldg_E_x, ldg_E_y               = get_static_variables(mesh.cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E,
          nflx_rho_x, nflx_rho_y,
          nflx_qx_x, nflx_qx_y,
          nflx_qy_x, nflx_qy_y,
          nflx_E_x, nflx_E_y             = get_bdry_variables(mesh.cache)

  @unpack bdry_rho, bdry_qx, bdry_qy, bdry_E,
          bdry_v,
          bdry_ldg_rho_x, bdry_ldg_rho_y,
          bdry_ldg_qx_x, bdry_ldg_qx_y,
          bdry_ldg_qy_x, bdry_ldg_qy_y,
          bdry_ldg_E_x, bdry_ldg_E_y,
          bdry_smoothed_mu               = get_bdry_variables(mesh.cache)
  @unpack v, smoothed_mu                 = get_static_variables(mesh.cache)

  ## TODO Maybe implement a macro with this syntax
  # @weakform(mesh) ∂t ldg_rho_x + ∂x rho = [ nflx_rho_x ]
  # -> compute_rhs_weak_form(ldg_rho_x, rho, 0, nflx_rho_x, mesh)
  # @weakform(mesh) ∂t ldg_rho_y + ∂y rho = [ nflx_rho_y ]
  # @weakform(mesh) ∂t ldg_qx_x  + ∂x qx  = [ nflx_qx_x ]
  # @weakform(mesh) ∂t ldg_qx_y  + ∂y qx  = [ nflx_qx_y ]
  # @weakform(mesh) ∂t ldg_qy_x  + ∂x qy  = [ nflx_qy_x ]
  # @weakform(mesh) ∂t ldg_qy_y  + ∂y qy  = [ nflx_qy_y ]
  # @weakform(mesh) ∂t ldg_E_x   + ∂x E   = [ nflx_E_x ]
  # @weakform(mesh) ∂t ldg_E_y   + ∂y E   = [ nflx_E_y ]

  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, qx, bdry_qx)
  dg1d.interpolate_face_data!(mesh, qy, bdry_qy)
  dg1d.interpolate_face_data!(mesh, E, bdry_E)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces!(ldg_nflux, equation, mesh)

  # TODO Need to revise Bdry conditions

  compute_rhs_weak_form!(ldg_rho_x, rho, 0,   nflx_rho_x, mesh)
  compute_rhs_weak_form!(ldg_rho_y, 0,   rho, nflx_rho_y, mesh)
  compute_rhs_weak_form!(ldg_qx_x,  qx,  0,   nflx_qx_x,  mesh)
  compute_rhs_weak_form!(ldg_qx_y,  0,   qx,  nflx_qx_y,  mesh)
  compute_rhs_weak_form!(ldg_qy_x,  qy,  0,   nflx_qy_x,  mesh)
  compute_rhs_weak_form!(ldg_qy_y,  0,   qy,  nflx_qy_y,  mesh)
  compute_rhs_weak_form!(ldg_E_x,   E,   0,   nflx_E_x,   mesh)
  compute_rhs_weak_form!(ldg_E_y,   0,   E,   nflx_E_y,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(flux, equation, mesh)
  broadcast_faces!(llf, equation, mesh)

  # TODO Need to revise Bdry conditions

  broadcast_volume!(mono_av_flux_2d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rho_x, bdry_ldg_rho_x)
  dg1d.interpolate_face_data!(mesh, ldg_rho_y, bdry_ldg_rho_y)
  dg1d.interpolate_face_data!(mesh, ldg_qx_x, bdry_ldg_qx_x)
  dg1d.interpolate_face_data!(mesh, ldg_qx_y, bdry_ldg_qx_y)
  dg1d.interpolate_face_data!(mesh, ldg_qy_x, bdry_ldg_qy_x)
  dg1d.interpolate_face_data!(mesh, ldg_qy_y, bdry_ldg_qy_y)
  dg1d.interpolate_face_data!(mesh, ldg_E_x, bdry_ldg_E_x)
  dg1d.interpolate_face_data!(mesh, ldg_E_y, bdry_ldg_E_y)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)
  broadcast_faces!(mono_av_nflux_2d, equation, mesh)

  # TODO don't we also need av_bdrylff here?

  compute_rhs_weak_form!(rhs_rho, flx_rho_x, flx_rho_y, nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_qx,  flx_qx_x,  flx_qx_y,  nflx_qx,  mesh)
  compute_rhs_weak_form!(rhs_qy,  flx_qy_x,  flx_qy_y,  nflx_qy,  mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E_x,   flx_E_y,   nflx_E,   mesh)

  return
end
