#######################################################################
#                              Timestep                               #
#######################################################################

function compute_maxspeed(::Project{:valencia1d}, equation, mesh)
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)
end
function compute_maxspeed(::Project{:spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)
end
function compute_maxspeed(::Project{:rescaled_spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
end
function compute_maxspeed(::Project{:doublecartoon}, equation, mesh)
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)
end
function compute_maxspeed(::Project{:cartoon}, equation, mesh)
  broadcast_volume!(maxspeed_cartoon, equation, mesh)
end

function get_maxspeed(mesh, equation)
  @unpack max_v = get_static_variables(mesh.cache)
  vmax = maximum(abs, max_v)
  vmax_limit = 0.99999999999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed $vmax exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end

function timestep(mesh, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  dl = min_grid_spacing(mesh)
  dt = dl / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = 1

function timestep(mesh::Mesh2d{SpectralElement}, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  @unpack element = mesh
  @unpack N, z = element
  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz
  dt = dl / (N * vmax)
  return dt
end


function timestep(mesh, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack equation = P

  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)

  # @unpack mu = get_cell_variables(mesh)
  # mumax = maximum(abs, mu)
  @unpack smoothed_mu = get_static_variables(mesh)
  mumax = maximum(abs, smoothed_mu)

  @unpack element = mesh
  @unpack N = element
  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K
  dl = min_grid_spacing(mesh)

  dt1 = dl / (vmax * N)
  dt2 = dL^2 / (mumax * N^3)
  dt = min(dt1, dt2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                               FV rhs                                #
#######################################################################


function rhs!(mesh::Mesh1d{FVElement}, P::Project)
  if P.prms.hrsc === :fv_central
    rhs_fv_central!(P, mesh)
  elseif P.prms.hrsc === :slope_limiter
    rhs_fv_muscl!(P, mesh)
  else
    TODO(P.prms.hrsc)
  end
end

function rhs_fv_central!(P::Project{:spherical1d}, mesh::Mesh1d{FVElement})

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                = get_dynamic_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, ρ, ϵ, vr, p,
          src_D, src_Sr, src_τ    = get_static_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ    = get_rhs_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)
  @unpack dt                      = get_global_variables(cache)

  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_D,  D,  flr_D,  src_D,  bdry_D,  nflr_D,  dt[1], mesh)
  fv_update_step!(rhs_Sr, Sr, flr_Sr, src_Sr, bdry_Sr, nflr_Sr, dt[1], mesh)
  fv_update_step!(rhs_τ,  τ,  flr_τ,  src_τ,  bdry_τ,  nflr_τ,  dt[1], mesh)

  return
end


function rhs_fv_central!(P::Project{:rescaled_spherical1d}, mesh::Mesh1d{FVElement})

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                = get_dynamic_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          D, Sr, τ,
          src_rD, src_rSr, src_rτ    = get_static_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ    = get_rhs_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_rescaled_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux_rescaled_spherical1d, equation, mesh)

  dt = timestep(mesh, P, hrsc)
  fv_update_step!(rhs_rD,  rD,  flr_rD,  src_rD,  bdry_D,  nflr_rD,  dt, mesh)
  fv_update_step!(rhs_rSr, rSr, flr_rSr, src_rSr, bdry_Sr, nflr_rSr, dt, mesh)
  fv_update_step!(rhs_rτ,  rτ,  flr_rτ,  src_rτ,  bdry_τ,  nflr_rτ,  dt, mesh)

  return
end


function rhs_fv_muscl!(P::Project, mesh::Mesh1d{FVElement})
  @unpack equation              = P
  @unpack D, Sr, τ             = get_dynamic_variables(mesh)
  @unpack D1, Sr1, τ1,
          D2, Sr2, τ2          = get_static_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ = get_rhs_variables(mesh)
  @unpack dt = get_global_variables(mesh)
  dt = dt[1]

  # SSPRK3 stepping from Hesthaven 2018 book, sec 10.2 (cf Script 10.13)
  step_fv_muscl!(P,mesh,(D,Sr,τ))
  @. D1  = D  + dt*rhs_D
  @. Sr1 = Sr + dt*rhs_Sr
  @. τ1  = τ  + dt*rhs_τ
  step_fv_muscl!(P,mesh,(D1,Sr1,τ1))
  @. D2  = (3*D  + D1  + dt*rhs_D)/4
  @. Sr2 = (3*Sr + Sr1 + dt*rhs_Sr)/4
  @. τ2  = (3*τ  + τ1  + dt*rhs_τ)/4
  step_fv_muscl!(P,mesh,(D2,Sr2,τ2))
  # atm final result is published via rhs_ variables
  @. rhs_D  = (D  + 2*D2  + 2*dt*rhs_D)/3
  @. rhs_Sr = (Sr + 2*Sr2 + 2*dt*rhs_Sr)/3
  @. rhs_τ  = (τ  + 2*τ2  + 2*dt*rhs_τ)/3

  return
end


function step_fv_muscl!(P::Project, mesh::Mesh1d{FVElement}, state)

  @unpack equation                      = P
  @unpack D, Sr, τ                      = get_dynamic_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ          = get_rhs_variables(mesh)
  @unpack bD, bSr, bτ,
          ρ, vr, ϵ,
          ρL, vrL, ϵL, pL,
          ρR, vrR, ϵR, pR,
          dρ_p, dvr_p, dϵ_p,
          dρ_m, dvr_m, dϵ_m,
          nflx_D_L, nflx_Sr_L, nflx_τ_L,
          nflx_D_R, nflx_Sr_R, nflx_τ_R,
          src_D, src_Sr, src_τ,
          γrr, α, βru                   = get_static_variables(mesh)
  @unpack eos                           = equation

  # this is a hack
  # the problem is that all @with_signature methods are tied to the
  # dynamic_variables of mesh.cache (D,Sr,τ)
  # but because we don't use the ordinary evolve interface, we have
  # to handle the state vector here manually
  # this requires us to first backup the current dynamic_variables,
  # copy the new state vector in there, compute stuff,
  # and then restore the dynamic_variables
  sD, sSr, sτ = state
  bD .= D
  bSr .= Sr
  bτ .= τ
  D .= sD
  Sr .= sSr
  τ .= sτ
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(flux_source_spherical1d, P.equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, P.equation, mesh)
  D .= bD
  Sr .= bSr
  τ .= bτ

  periodic = isperiodic(mesh)
  prms_slope_limiter = (; method=P.prms.slope_limiter_method,
                          tvb_M=P.prms.slope_limiter_tvb_M)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)
  dl = L/K

  for vars in ( (ρL,ρR,ρ,dρ_m,dρ_p),
                (vrL,vrR,vr,dvr_m,dvr_p),
                (ϵL,ϵR,ϵ,dϵ_m,dϵ_p) )

    wL,wR,w,dw₋,dw₊ = vars

    # compute left (-) and right (+) differences
    @turbo for k in 2:K-1
      k₋, k₊ = k-1, k+1
      w₋, w₀, w₊ = w[k₋], w[k], w[k₊]
      dw₊[k], dw₋[k] = w₊-w₀, w₀-w₋
    end

    # limit slope and compute left (L) and (R) interface values
    for k in 1:K
      dwL = limit_slope(dw₊[k], dw₋[k], dl, prms_slope_limiter)
      wL[k] = w[k] - dwL/2
      wR[k] = w[k] + dwL/2
    end
  end

  for k in 1:K
    pL[k] = eos(Pressure, Density, ρL[k], InternalEnergy, ϵL[k])
    pR[k] = eos(Pressure, Density, ρR[k], InternalEnergy, ϵR[k])
  end

  # numerical fluxes
  # subscript + ... cell interior value
  # subscript - ... cell exterior value
  # nx ... pointing from interior too exterior cell
  @unpack max_v    = get_static_variables(mesh)
  for k in 2:K
    # on left cell interfaces
    nx             = -1
    k₊, k₋         = k, k-1
    γrr₊, γrr₋     = γrr[k₊], γrr[k₋]
    γuurr₊, γuurr₋ = 1/γrr₊, 1/γrr₋
    vr₊, vr₋       = vrL[k₊], vrR[k₋]
    ρ₊, ρ₋         = ρL[k₊], ρR[k₋]
    ϵ₊, ϵ₋         = ϵL[k₊], ϵR[k₋]
    p₊, p₋         = pL[k₊], pR[k₋]
    vru₊, vru₋     = vr₊*γuurr₊, vr₋*γuurr₋
    W₊, W₋         = 1/sqrt(1-vru₊*vr₊), 1/sqrt(1-vru₋*vr₋)
    D₊, D₋         = ρ₊*W₊, ρ₋*W₋
    h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
    Sr₊, Sr₋       = ρ₊*h₊*W₊^2*vr₊, ρ₋*h₋*W₋^2*vr₋
    τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
    α₊, α₋         = α[k₊], α[k₋]
    βru₊, βru₋     = βru[k₊], βru[k₋]
    vv             = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋         = α₊*D₊*(vru₊-βru₊/α₊), α₋*D₋*(vru₋-βru₋/α₋)
    nflx_D_L[k]    = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋         = α₊*(Sr₊*(vru₊-βru₊/α₊)+p₊), α₋*(Sr₋*(vru₋-βru₋/α₋)+p₋)
    nflx_Sr_L[k]   = LLF(nx*f₊, nx*f₋, Sr₊, Sr₋, vv)
    f₊, f₋         = α₊*(τ₊*(vru₊-βru₊/α₊)+p₊*vru₊), α₋*(τ₋*(vru₋-βru₋/α₋)+p₋*vru₋)
    nflx_τ_L[k]    = LLF(nx*f₊, nx*f₋, τ₊, τ₋, vv)
  end

  for k in 1:K-1
    # on right cell interfaces
    nx            = 1
    k₊, k₋        = k, k+1
    γrr₊, γrr₋     = γrr[k₊], γrr[k₋]
    γuurr₊, γuurr₋ = 1/γrr₊, 1/γrr₋
    vr₊, vr₋       = vrR[k₊], vrL[k₋]
    ρ₊, ρ₋         = ρR[k₊], ρL[k₋]
    ϵ₊, ϵ₋         = ϵR[k₊], ϵL[k₋]
    p₊, p₋         = pR[k₊], pL[k₋]
    vru₊, vru₋     = vr₊*γuurr₊, vr₋*γuurr₋
    W₊, W₋         = 1/sqrt(1-vru₊*vr₊), 1/sqrt(1-vru₋*vr₋)
    D₊, D₋         = ρ₊*W₊, ρ₋*W₋
    h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
    Sr₊, Sr₋       = ρ₊*h₊*W₊^2*vr₊, ρ₋*h₋*W₋^2*vr₋
    τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
    α₊, α₋         = α[k₊], α[k₋]
    βru₊, βru₋     = βru[k₊], βru[k₋]
    vv             = absmax(max_v[k₊], max_v[k₋])
    f₊, f₋         = α₊*D₊*(vru₊-βru₊/α₊), α₋*D₋*(vru₋-βru₋/α₋)
    nflx_D_R[k]    = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
    f₊, f₋         = α₊*(Sr₊*(vru₊-βru₊/α₊)+p₊), α₋*(Sr₋*(vru₋-βru₋/α₋)+p₋)
    nflx_Sr_R[k]   = LLF(nx*f₊, nx*f₋, Sr₊, Sr₋, vv)
    f₊, f₋         = α₊*(τ₊*(vru₊-βru₊/α₊)+p₊*vru₊), α₋*(τ₋*(vru₋-βru₋/α₋)+p₋*vru₋)
    nflx_τ_R[k]    = LLF(nx*f₊, nx*f₋, τ₊, τ₋, vv)
  end

  # evaluate rhs
  @turbo for k in 1:K
    # we have a + in the braces, because we use the nx with fluxes above
    rhs_D[k]  = -(nflx_D_R[k]  + nflx_D_L[k] )/dl + src_D[k]
    rhs_Sr[k] = -(nflx_Sr_R[k] + nflx_Sr_L[k])/dl + src_Sr[k]
    rhs_τ[k]  = -(nflx_τ_R[k]  + nflx_τ_L[k] )/dl + src_τ[k]
  end

  # because we don't have bdry conditions atm we cheat here
  rhs_D[1]  = rhs_D[end]  = 0
  rhs_Sr[1] = rhs_Sr[end] = 0
  rhs_τ[1]  = rhs_τ[end]  = 0

  return
end


@inline function limit_slope(a, b, dx, prms)
  if prms.method === :none
    return 0.0
  elseif prms.method === :minmod
    return TCI.minmod(a, b)
  elseif prms.method === :muscl
    return TCI.minmod((a+b)/2, 2*a, 2*b)
  elseif prms.method === :superbee
    return TCI.minmod(TCI.maxmod(a,b),TCI.minmod(2*a,2*b))
  elseif prms.method === :van_albada
    c = dx^3
    return TCI.minmod( ((a^2+c^2)*b+(b^2+c^2)*a)/(a^2+b^2+c^2), 2*a, 2*b )
  elseif prms.method === :van_leer
    return TCI.minmod(2*a*b/(a+b), 2*a, 2*b)
  elseif prms.method === :tvb
    return TCI.minmod_M(dx, prms.tvb_M, (a+b)/2, 2*a, 2*b)
  else
    TODO(prms.method)
  end
end


#######################################################################
#                               DG RHS                                #
#######################################################################


rhs!(mesh::Mesh1d, P::Project) = rhs!(mesh, P, P.hrsc)


function rhs!(mesh::Mesh1d, P::Project{:valencia1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                 = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ     = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          D, Sr, τ, max_v, vr, p,
          src_rD, src_rSr, src_rτ     = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_valencia1d, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_valencia1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_valencia1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    if ismirrored(mesh)
      broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
    else
      impose_symmetry_bdry_values_valencia1d!(mesh)
      broadcast_bdry!(bdryllf_valencia1d_tov_reflected, equation, mesh)
    end
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  return
end


function rhs!(mesh::Mesh1d, P::Project{:spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                 = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ     = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, vr, p,
          src_D, src_Sr, src_τ     = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_spherical1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    if ismirrored(mesh)
      broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
    else
      impose_symmetry_bdry_values_spherical1d!(mesh)
      broadcast_bdry!(bdryllf_spherical1d_tov_reflected, equation, mesh)
    end
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  return
end


function rhs!(mesh::Mesh1d, P::Project{:rescaled_spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                    = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ        = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          max_v, vr, p, D, Sr, τ,
          src_rD, src_rSr, src_rτ        = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_rescaled_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_faces!(llf_rescaled_spherical1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_rescaled_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_rescaled_spherical1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  return
end


function rhs!(mesh::Mesh1d, P::Project{:doublecartoon}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rτ                 = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rτ     = get_rhs_variables(cache)
  @unpack flx_rDx, flx_rSxx, flx_rτx,
          D, Sx, τ, max_v, vx, p,
          src_rD, src_rSx, src_rτ     = get_static_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rτ,
          bdry_D, bdry_Sx, bdry_τ,
          bdry_max_v, bdry_vx, bdry_p = get_bdry_variables(cache)

  # impose_symmetry_state!(P, mesh)
  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, equation, mesh)
  broadcast_faces!(llf_doublecartoon, equation, mesh)
  if ismirrored(mesh)
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi, equation, mesh)
    elseif P.prms.problem === :tov
      broadcast_bdry!(bdryllf_doublecartoon_tov, equation, mesh)
    else
      TODO()
    end
  else
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi_mirrored, equation, mesh)
    elseif P.prms.problem === :tov
      impose_symmetry_bdry_values_doublecartoon!(mesh)
      broadcast_bdry!(bdryllf_doublecartoon_tov_mirrored, equation, mesh)
    else
      TODO()
    end
  end

  compute_rhs_weak_form!(rhs_rD,  flx_rDx,  src_rD,  nflx_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSx, flx_rSxx, src_rSx, nflx_rSx, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  src_rτ,  nflx_rτ,  mesh)

  return
end


function rhs!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractReconstruction)

  @unpack equation = P

  @unpack D, Sr, τ                    = get_dynamic_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ        = get_rhs_variables(mesh)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, vr, p,
          src_D, src_Sr, src_τ        = get_static_variables(mesh)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(mesh)
  @unpack flag                        = get_cell_variables(mesh)

  ismirrored(mesh) || TODO("only tested for mirrored domain")

  HRSC.reconstruct!(D,  flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  HRSC.reconstruct!(Sr, flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  HRSC.reconstruct!(τ,  flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)

  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_spherical1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  return
end


function rhs!(mesh, P::Project{:valencia1d}, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ             = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          ldg_rD, ldg_rSr, ldg_rτ,
          D, Sr, τ,
          max_v, vr, p, smoothed_mu,
          src_rD, src_rSr, src_rτ = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_ldg_rD, bdry_ldg_rSr, bdry_ldg_rτ,
          bdry_rhs_rD, bdry_rhs_rSr, bdry_rhs_rτ,
          bdry_smoothed_mu        = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_valencia1d, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_valencia1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_valencia1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_valencia1d, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_valencia1d, equation, mesh)
  compute_rhs_weak_form!(ldg_rD,  rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(ldg_rSr, rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(ldg_rτ,  rτ,  nflr_rτ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_{rD,rSr,rtau}
  # numerical fluxes
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
  else
    TODO()
  end

  broadcast_volume!(av_flux_covariant_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rD,       bdry_ldg_rD)
  dg1d.interpolate_face_data!(mesh, ldg_rSr,      bdry_ldg_rSr)
  dg1d.interpolate_face_data!(mesh, ldg_rτ,       bdry_ldg_rτ)
  dg1d.interpolate_face_data!(mesh, rhs_rD,       bdry_rhs_rD)
  dg1d.interpolate_face_data!(mesh, rhs_rSr,      bdry_rhs_rSr)
  dg1d.interpolate_face_data!(mesh, rhs_rτ,       bdry_rhs_rτ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_valencia1d, equation, mesh)

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  return
end


function rhs!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                            = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ                = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          ldg_D, ldg_Sr, ldg_τ,
          max_v, vr, p, smoothed_mu,
          src_D, src_Sr, src_τ                = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_ldg_D, bdry_ldg_Sr, bdry_ldg_τ,
          bdry_rhs_D, bdry_rhs_Sr, bdry_rhs_τ,
          bdry_smoothed_mu                    = get_bdry_variables(cache)

  ismirrored(mesh) || TODO("only tested for mirrored domain")

  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_spherical1d, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_spherical1d, equation, mesh)
  compute_rhs_weak_form!(ldg_D,  D,  nflr_D,  mesh)
  compute_rhs_weak_form!(ldg_Sr, Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(ldg_τ,  τ,  nflr_τ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_{D,Sr,tau}
  # numerical fluxes
  broadcast_faces!(llf, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
  else
    TODO()
  end

  broadcast_volume!(av_flux_covariant_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D,       bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_Sr,      bdry_ldg_Sr)
  dg1d.interpolate_face_data!(mesh, ldg_τ,       bdry_ldg_τ)
  dg1d.interpolate_face_data!(mesh, rhs_D,       bdry_rhs_D)
  dg1d.interpolate_face_data!(mesh, rhs_Sr,      bdry_rhs_Sr)
  dg1d.interpolate_face_data!(mesh, rhs_τ,       bdry_rhs_τ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_spherical1d, equation, mesh)

  compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)

  return
end


#######################################################################
#                               Mesh2d                                #
#######################################################################


function rhs!(mesh::Mesh2d, P::Project{:cartoon})

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rSz, rτ                      = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rSz, rhs_rτ      = get_rhs_variables(cache)
  @unpack flx_rDx, flx_rDz,
          flx_rSxx, flx_rSxz,
          flx_rSzx, flx_rSzz,
          flx_rτx, flx_rτz,
          D, Sx, Sz, τ, max_v, vx, vz, p,
          src_rD, src_rSx, src_rSz, src_rτ      = get_static_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rSz, nflx_rτ,
          bdry_D, bdry_Sx, bdry_Sz, bdry_τ,
          bdry_max_v,  bdry_vx, bdry_vz, bdry_p = get_bdry_variables(cache)

  # impose_symmetry_state!(P, mesh)
  broadcast_volume!(unscale_conservatives_cartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm
    broadcast_volume!(cons2prim_cartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_cartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_cartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, Sz,    bdry_Sz)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, vz,    bdry_vz)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_cartoon!(P, mesh)
  broadcast_volume!(flux_source_cartoon, equation, mesh)
  broadcast_faces!(llf_cartoon, equation, mesh)
  # assuming reflection symmetry for now
  impose_symmetry_bdry_values_cartoon!(mesh)
  broadcast_bdry!(bdryllf_cartoon_tov_reflected, equation, mesh)

  compute_rhs_weak_form!(rhs_rD,  flx_rDx,  flx_rDz,  src_rD,  nflx_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSx, flx_rSxx, flx_rSxz, src_rSx, nflx_rSx, mesh)
  compute_rhs_weak_form!(rhs_rSz, flx_rSzx, flx_rSzz, src_rSz, nflx_rSz, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  flx_rτz,  src_rτ,  nflx_rτ,  mesh)

  return
end
