"""
    @with_signature

Macro to extract `accepted` and `returned` arguments from a function and make them
programmatically accessible. Functions defined using this macro can be used with
the `broadcast_[volume|faces|bdrys]_2!` interfaces.

To test a `@with_signature` in standalone mode, see below for an example.

The following options are available to customize `@with_signature` definitions:
- `skip_debug_info::Bool=false`: If `true` the generated code does not contain
  LineNumberNodes which makes it such that when an error occurs in the computation, Julia can't tell
  where it came from in the definition of the function. Only available to easily debug
  if the function parsing works correctly.
- `inbounds::Bool=true`: Controls `@inbounds` access for the grid loop.
- `simd::Bool=true`: Controls `@simd` optimization for the grid loop.
  Use `simd=false` for debugging to restore correct line numbers in stack traces.
- `debug::Bool=false`: To disable `inbounds, simd` optimizations.
- `print_idxs=false`: To enable printing of volume and boundary indices.

The following options are available to customize the keyword `@accepts, @returns` definitions:
- `Prefix(name)`: Append `name_` to all names following the keywords.
- `bdry`: Indicate that data from the variables following this option should be loaded
from the boundary indices. Default is to load data from volume indices.


# Example
```
@with_signature function flux(eq::Equation, mesh::Mesh1d)
   @accepts fx, fy, u, v
   @accepts [Prefix(out),bdry] fx, fy, u, v
   @accepts [bdry] nx, ny
   nflx     = nx*fx + ny*fy
   out_nflx = nx*out_fx + ny*out_fy
   nflx     = LLF(nflx, out_nflx, u, out_u, absmax(v, out_v))
   @returns [bdry] nflx
end
```
translates (roughly) into
```
function flux(eq::Equation, mesh::Mesh1d, ::Any, cache, arguments, returns, volumeidxs, bdryidxs)

  # unpack
  _arg_fx     = get_variable(cache, :fx)
  _arg_fy     = get_variable(cache, :fy)
  _arg_u      = get_variable(cache, :u)
  _arg_v      = get_variable(cache, :v)
  _arg_out_fx = get_variable(cache, :out_fx)
  _arg_out_fy = get_variable(cache, :out_fy)
  _arg_out_u  = get_variable(cache, :out_u)
  _arg_out_v  = get_variable(cache, :out_v)
  _arg_out_nx = get_variable(cache, :out_nx)
  _arg_out_ny = get_variable(cache, :out_ny)
  _ret_nflx   = get_variable(cache, :out_nflx)

  @inbounds for ii in eachindex(volumeidxs, bdryidxs)
    vi = volumeidxs[ii]
    bi = bdryidxs[ii]

    # read
    fx     = _arg_fx[vi]
    fy     = _arg_fy[vi]
    u      = _arg_u[vi]
    v      = _arg_v[vi]
    out_fx = _arg_out_fx[bi]
    out_fy = _arg_out_fy[bi]
    out_u  = _arg_out_u[bi]
    out_v  = _arg_out_v[bi]
    nx     = _arg_out_nx[bi]
    ny     = _arg_out_ny[bi]

    # actual computation
    nflx     = nx*fx + ny*fy
    out_nflx = nx*out_fx + ny*out_fy
    nflx     = LLF(nflx, out_nflx, u, out_u, absmax(v, out_v))

    # write
    _ret_nflx[bi] = nflx
  end
end

# helpers
accepts(::typeof(fn), ::Equation, mesh::Mesh1d) = (:fx, :fy, :u, :v, :fx, :fy, :u, :v, :ny, :ny)
returns(::typeof(fn), ::Equation, mesh::Mesh1d) = (:nflx,)
has_signature(::typeof(fn), ::Equation, mesh::Mesh1d) = true
```

The above can be called as
```
N = 100
args = [ rand(N) for _ = 1:11 ]
rets = [ rand(N) for _ = 1:1  ]
volumeidxs = 1:100
bdryidxs   = 100:-1:1
eq = Equation()
mesh = Mesh1d(N=1,K=100)
flux(eq, args, rets, volumeidxs, bdryidxs)
```
"""
macro with_signature(ex)
  eex = make_with_signature_methods(ex, parse_options())
  return esc(eex)
end
macro with_signature(ex_option, ex)
  options = parse_options(ex_option)
  eex = make_with_signature_methods(ex, options)
  return esc(eex)
end


const _empty_opts = (quote [] end).args[2]
function parse_options(ex=_empty_opts)
  if !_is_vect(ex)
    error("@with_signature: invalid options '$ex', try instead '@with_signature [option_1=true] function flux(eq::Equation) ... end'")
  end
  options = Dict{Symbol,Any}()
  args = ex.args
  for a in args
    matched = MacroTools.@capture(a, lhs_ = rhs_)
    if !matched
      error("@with_signature: invalid options '$ex', try instead '@with_signature [option_1=true] function flux(eq::Equation) ... end'")
    end
    if haskey(options, lhs)
      error("@with_signature: duplicated option found in '$ex'")
    end
    if lhs === :skip_debug_info
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'skip_debug_info', expected a Bool")
      end
      options[:skip_debug_info] = rhs
      continue
    elseif lhs === :print_idxs
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'print_idxs', expected a Bool")
      end
      options[:print_idxs] = rhs
      continue
    elseif lhs === :inbounds
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'inbounds', expected a Bool")
      end
      options[:inbounds] = rhs
      continue
    elseif lhs === :simd
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'simd', expected a Bool")
      end
      options[:simd] = rhs
      continue
    elseif lhs === :debug
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'debug', expected a Bool")
      end
      options[:debug] = rhs
      continue
    end
    error("@with_signature: unknown option '$a' found in '$ex', skipping ...")
  end
  # set defaults
  get!(options, :skip_debug_info, false)
  get!(options, :print_idxs, false)
  get!(options, :inbounds, true)
  get!(options, :simd, true)
  get!(options, :debug, false)
  if options[:debug]
    options[:inbounds] = false
    options[:simd]     = false
  end
  return options
end


"""
See @with_signature
"""
macro accepts(ex); ex; end
macro returns(ex); ex; end


_is_tuple(ex) = ex isa Expr && ex.head === :tuple
_is_vect(ex) = ex isa Expr && ex.head === :vect


# parses a `ex = :([option=1,option2] a, b, c, d)` and returns
# `Dict(option=>1, option=>2], (:c, :d, :e, :f)`
#
# `keyword, line` are only for error messages
function parse_options_variables(keyword, line, ex)

  # unpack
  ex_options, ex_args = if length(ex) == 1
    nothing, ex[1]
  elseif length(ex) == 2
    ex[1], ex[2]
  else
    error("$keyword: invalid syntax in '$line', use as '$keyword [options] a, b, c'")
  end

  # unpack options
  options = Dict{Symbol,Any}()
  if !isnothing(ex_options)
    if !_is_vect(ex_options)
      error("$keyword: invalid syntax for options '$ex_options' found in '$ex', use as '$keyword [options] a, b, c'")
    end
    args = ex_options.args
    for a in args
      if haskey(options, a)
        error("$keyword: duplicated option found in '$ex'")
      end
      if a === :bdry
        options[:bdry] = true
        continue
      end
      matched = MacroTools.@capture(a, Prefix(match_))
      if matched
        if haskey(options, :prefix)
          error("$keyword: duplicated option found in '$ex'")
        end
        options[:prefix] = match
        continue
      end
      error("$keyword: unknown option '$a' found in '$ex_options', skipping ...")
    end
  end
  # set defaults
  get!(options, :bdry, false)
  get!(options, :prefix, nothing)

  vars = if ex_args isa Symbol
    [ex_args]
  elseif _is_tuple(ex_args)
    vars = ex_args.args[1:end]
    for v in vars
      if !(v isa Symbol)
        error("$keyword: invalid variable '$v' found in '$line', only use Symbols with $keyword")
      end
    end
    vars
  else
    error("$keyword: invalid syntax for variables found in '$line', use as '$keyword [options] a, b, c'")
  end

  options, vars
end


parse_accepts(line, ex) = parse_options_variables("@accepts", line, ex)
parse_returns(line, ex) = parse_options_variables("@returns", line, ex)


function make_with_signature_methods(expr, options=(;))
  # display(expr)
  ex = MacroTools.prewalk(MacroTools.rmlines, expr)
  # display(ex)

  ex.head === :function || error("@with_signature: only works with functions")
  fn_signature = ex.args[1]
  2 <= length(fn_signature.args) <= 3 || error(
    "@with_signature: function with exactly one or two arguments required")
  if length(fn_signature.args) == 2
    # insert a dummy argument
    push!(fn_signature.args, :(mesh::Any))
  end
  fn_name, first_arg, second_arg = fn_signature.args
  lno_func = expr.args[2].args[1] # line number node

  # sanitize function arguments
  (first_arg isa Expr && first_arg.head === :(::)) || error(
    "@with_signature: function argument '$first_arg' must be type annotated")
  dispatch_type_first = last(first_arg.args)
  dispatch_type_second = if second_arg isa Symbol
    :Any
  elseif second_arg isa Expr && second_arg.head === :(::)
    last(second_arg.args)
  else
    error("@with_signature: invalid argument '$second_arg'")
  end

  ex.args[2].head === :block || error("@with_singature: failed to extract function body")
  fn_body = ex.args[2].args

  # extract place holders
  length(fn_body) == 0 && error("@with_signature: missing function body")
  accepts_lines, accepts_exs = Any[], Any[]
  for line in fn_body
    matched_accepts = MacroTools.@capture(line, @accepts accepts_ex__)
    !matched_accepts && break
    push!(accepts_lines, line)
    push!(accepts_exs, accepts_ex)
  end
  length(accepts_lines) == 0 && error(
    "@with_signature: expected '@accepts [options] a, b, ...' at the start of the function body")
  lno_accepts = [ expr.args[2].args[2+2*(i-1)] for i in 1:length(accepts_lines) ] # line number nodes
  # @show lno_accepts

  returns_lines, returns_exs = Any[], Any[]
  for line in reverse(fn_body) # search for @returns from bottom to top
    matched_returns = MacroTools.@capture(line, @returns returns_ex__)
    !matched_returns && break
    push!(returns_lines, line)
    push!(returns_exs, returns_ex)
  end
  length(returns_lines) == 0 && error(
    "@with_signature: expected '@returns [options] a, b, ...' at the end of the function body")
  # reverse returns expressions so that we emmit code from top to bottom
  returns_exs   = returns_exs[end:-1:1]
  returns_lines = returns_lines[end:-1:1]
  lno_returns = [ expr.args[2].args[end-1-2*(i-1)] for i in 1:length(returns_lines) ] # line number nodes
  # @show lno_returns

  # @accepts and @returns can only occur right at the start or end
  if any(line -> MacroTools.@capture(line, @accepts _), fn_body[length(accepts_lines)+1:end])
    error("@with_signature: expected '@accepts [options] a, b, ...' at the start of the function body")
  end
  if any(line -> MacroTools.@capture(line, @returns _), fn_body[1:end-length(returns_lines)])
    error("@with_signature: expected '@returns [options] a, b, ...' at the end of the function body")
  end

  # gather variables from place holders
  accepts_options_vars = parse_accepts.(accepts_lines, accepts_exs)
  returns_options_vars = parse_returns.(returns_lines, returns_exs)

  # exctract compute body from initial expression in order to preserve
  # line number nodes for debugging
  compute_body = if !options[:skip_debug_info]
    try
      @assert expr.head === :function
      init_fn_body = expr.args[2].args
      init_fn_body[2+2*length(accepts_lines):end-2*length(returns_lines)]
    catch
      @warn "@with_signature: failed to extract LineNumberNodes for function body"
      fn_body[1+length(accepts_lines):end-length(returns_lines)]
    end
  else
      fn_body[1+length(accepts_lines):end-length(returns_lines)]
  end
  if length(compute_body) == 0
    error("""@with_signature: nothing to compute in expression
            @with_signature $ex
          """)
  end

  # make gen'ed symbols for construction of the rest of the function body
  ii, vi, bi, vidxs, bidxs =
    gensym.((:ii,:vi,:bi,:vidxs,:bidxs))

  # for each variable in @accepts
  # 1) resolve the prefix,
  # 2) make a gen'ed symbol for array unpacking
  # 3) construct the array unpacking calls
  acc_vars, prefix_acc_vars, gen_acc_vars = Vector{Symbol}[], Vector{Symbol}[], Vector{Symbol}[]
  bdry_acc, unpack_acc = Vector{Bool}[], Union{LineNumberNode,Expr}[]
  n = 1
  for (i,((opts, vars),lno)) in enumerate(zip(accepts_options_vars,lno_accepts))
    push!(acc_vars, vars)
    push!(prefix_acc_vars, isnothing(opts[:prefix]) ? vars : Symbol.(opts[:prefix],:_,vars))
    push!(bdry_acc, fill(opts[:bdry], length(vars)))
    gens = Symbol[]
    push!(unpack_acc, lno)
    for v in vars
      gv = gensym(v)
      push!(unpack_acc, :($gv = get_variable(cache, $(QuoteNode(v)))))
      push!(gens, gv)
      n += 1
    end
    push!(gen_acc_vars, gens)
  end

  # for each variable in @returns
  # 1) resolve the prefix,
  # 2) make a gen'ed symbol for array unpacking
  # 3) construct the array unpacking calls
  ret_vars, prefix_ret_vars, gen_ret_vars = Vector{Symbol}[], Vector{Symbol}[], Vector{Symbol}[]
  bdry_ret, unpack_ret, =  Vector{Bool}[], Union{LineNumberNode,Expr}[]
  n = 1
  for (i,((opts, vars),lno)) in enumerate(zip(returns_options_vars,lno_returns))
    push!(ret_vars, vars)
    push!(prefix_ret_vars, isnothing(opts[:prefix]) ? vars : Symbol.(opts[:prefix],:_,vars))
    push!(bdry_ret, fill(opts[:bdry], length(vars)))
    gens = Symbol[]
    push!(unpack_ret, lno)
    for v in vars
      gv = gensym(v)
      push!(unpack_ret, :($gv = get_variable(cache, $(QuoteNode(v)))))
      push!(gens, gv)
      n += 1
    end
    push!(gen_ret_vars, gens)
  end

  # error on duplicated variable names (after resolving prefixes)
  red_prefix_acc_vars = reduce(vcat, prefix_acc_vars)
  if !allunique(prefix_acc_vars)
    uvars = unique(red_prefix_acc_vars)
    idxdups = findall(v -> count(Base.Fix1(===,v),red_prefix_acc_vars) > 1, uvars)
    dups = red_prefix_acc_vars[idxdups]
    error("duplicated arguments in all @accepts found (after resolving prefixes): $(join(dups,", "))")
  end
  red_prefix_ret_vars = reduce(vcat, prefix_ret_vars)
  if !allunique(red_prefix_ret_vars)
    uvars = unique(red_prefix_ret_vars)
    idxdups = findall(v -> count(Base.Fix1(===,v),red_prefix_ret_vars) > 1, uvars)
    dups = red_prefix_ret_vars[idxdups]
    error("duplicated arguments in all @returns found (after resolving prefixes): $(join(dups,", "))")
  end

  # construct array getter and setters for loop body, e.g.
  #   var"##vi#2802" = var"##vidxs#2804"[var"##ii#2801"]
  getvars_lines = [ [ :( $pv = $gv[$(bdry ? bi : vi)] ) for (pv,gv,bdry) in zip(prefixs,gens,bdrys) ]
                    for (prefixs,gens,bdrys) in zip(prefix_acc_vars,gen_acc_vars,bdry_acc) ]
  setvars_lines = [ [ :( $gv[$(bdry ? bi : vi)] = $pv ) for (pv,gv,bdry) in zip(prefixs,gens,bdrys) ]
                    for (prefixs,gens,bdrys) in zip(prefix_ret_vars,gen_ret_vars,bdry_ret) ]

  # loop skeleton
  loop_ex = if options[:print_idxs]
    quote
      for $ii in eachindex($vidxs, $bidxs)
        $vi = $vidxs[$ii]
        $bi = $bidxs[$ii]
        println((;ii=$ii, vi=$vi, bi=$bi))
      end
    end
  else
    quote
      for $ii in eachindex($vidxs, $bidxs)
        $vi = $vidxs[$ii]
        $bi = $bidxs[$ii]
      end
    end
  end
  # strip artificial LineNumberNodes and unwrap a begin block
  loop_ex = MacroTools.prewalk(MacroTools.rmlines, loop_ex)
  loop_ex = loop_ex.args[1]
  loop_body = loop_ex.args[2].args

  # insert getvars_lines with original LineNumberNodes
  for (lno,getvars) in zip(lno_accepts,getvars_lines)
    push!(loop_body, lno)
    append!(loop_body, getvars)
  end

  # contains original LineNumberNodes
  append!(loop_body, compute_body)

  # insert setvars_lines with original LineNumberNodes
  for (lno,setvars) in zip(lno_returns,setvars_lines)
    push!(loop_body, lno)
    append!(loop_body, setvars)
  end

  # prepend macros to for loop
  if options[:simd]
    loop_ex = Expr(:macrocall, Symbol("@simd"), nothing, loop_ex)
  end
  if options[:inbounds]
    loop_ex = Expr(:macrocall, Symbol("@inbounds"), nothing, loop_ex)
  end

  # combine loop with accepts, returns array unpacking
  new_func = deepcopy(ex)
  new_func.args[2].args = vcat(unpack_acc, unpack_ret, loop_ex)

  # adjust function signature
  new_fn_signature = Expr(:call, fn_name, first_arg, second_arg, :cache, vidxs, bidxs)
  new_func.args[1] = new_fn_signature

  # construct @with_signature interface functions signature(), hassignature()
  tpl_acc_vars = Tuple(reduce(vcat, acc_vars))
  tpl_ret_vars = Tuple(reduce(vcat, ret_vars))
  thismod = @__MODULE__

  ex_sig = quote
    $(thismod).signature(::typeof($(fn_name)), dispatch::$(dispatch_type_first), mesh::$(dispatch_type_second)) = ($(tpl_acc_vars), $(tpl_ret_vars))
  end
  # strip artificial LineNumberNodes and unwrap a begin block
  ex_sig = MacroTools.prewalk(MacroTools.rmlines, ex_sig)
  ex_sig = ex_sig.args[1]
  ex_hassig = quote
    $(thismod).has_signature(::typeof($(fn_name)), dispatch::Type{<:$(dispatch_type_first)}, mesh::$(dispatch_type_second)) = true
  end
  # strip artificial LineNumberNodes and unwrap a begin block
  ex_hassig = MacroTools.prewalk(MacroTools.rmlines, ex_hassig)
  ex_hassig = ex_hassig.args[1]

  # TODO What about docstrings for these methods?
  defs = Expr(:block, lno_func, new_func, lno_func, ex_sig, lno_func, ex_hassig)

  # display(MacroTools.prewalk(MacroTools.rmlines, new_ex))

  return defs
end


#######################################################################
#                          Interface methods                          #
#######################################################################


has_signature(fn, dispatch, mesh) = false
signature(fn, dispatch, mesh) = error(
  "function '$fn' with dispatch '$(typeof(dispatch))' was not defined using @with_signature")
accepts(fn, dispatch, mesh) = signature(fn, dispatch)[1]
returns(fn, dispatch, mesh) = signature(fn, dispatch)[2]

signature(fn, dispatch, mesh=nothing) = error(
  "function '$fn' with dispatch '$(typeof(dispatch))' and mesh '$(typeof(mesh))' was not defined using @with_signature")
has_signature(fn, dispatch::Type, mesh=nothing) = false
has_signature(fn, dispatch, any) = has_signature(fn, typeof(dispatch), nothing)


#######################################################################
#                          broadcast methods                          #
#######################################################################


function broadcast_computation!(f, equation, mesh, volidxs, bdryidxs)
  @toggled_assert has_signature(f, equation, mesh) "'$f(::$(typeof(equation)))' was not defined using @with_signature"
  @inline f(equation, mesh, mesh.cache, volidxs, bdryidxs)
  return
end

function broadcast_volume!(f, equation, mesh::AbstractMesh)
  # @timeit TO "broadcast_volume!" begin
    broadcast_computation!(f, equation, mesh, volumeindices(mesh)...)
  # end
end
function broadcast_faces!(f, equation, mesh::AbstractMesh)
  # @timeit TO "broadcast_faces!" begin
    broadcast_computation!(f, equation, mesh, faceindices(mesh)...)
  # end
end
function broadcast_bdry!(f, equation, mesh::AbstractMesh)
  # @timeit TO "broadcast_bdry!" begin
    broadcast_computation!(f, equation, mesh, bdryindices(mesh)...)
  # end
end
